@@ -26,24 +26,12 @@
 #ifndef GPAPASTRINGS_H
 #define GPAPASTRINGS_H
 
-#include <gpapa.h>
+#include <glib.h>
+#include <gpgme.h>
 
-gchar * gpa_keytrust_string (GpapaKeytrust keytrust);
-
-gchar * gpa_ownertrust_string (GpapaOwnertrust ownertrust);
-GpapaOwnertrust gpa_ownertrust_from_string (gchar * string);
-const char * gpa_ownertrust_icon_name (GpapaOwnertrust ownertrust);
-
-gchar * gpa_algorithm_string (GpapaAlgo algo);
-GpapaAlgo gpa_algorithm_from_string (gchar * string);
-
-gchar * gpa_unit_expiry_time_string(int index);
+gchar * gpa_unit_expiry_time_string(int idx);
 gchar gpa_time_unit_from_string (gchar * string);
-gchar * gpa_expiry_date_string (GDate * expiry_date);
-gchar * gpa_creation_date_string (GDate * creation_date);
-
-gchar * gpa_file_status_string (GpapaFileStatus status);
-
-gchar * gpa_sig_validity_string (GpapaSigValidity validity);
+gchar * gpa_expiry_date_string (unsigned long expiry_time);
+gchar * gpa_creation_date_string (unsigned long creation_time);
 
 #endif /* GPAPASTRINGS_H */
