@@ -21,6 +21,8 @@
 #include "gpgmeedit.h"
 #include "passwddlg.h"
 
+#include <unistd.h>
+
 /* The edit callback for all the edit operations is edit_fnc(). Each
  * operation is modelled as a sequential machine (a Moore machine, to be
  * precise). Therefore, for each operation you must write two functions.
@@ -35,12 +37,12 @@
  */
 
 /* Prototype of the action function. Returns the error if there is one */
-typedef GpgmeError (*EditAction) (int state, void *opaque,
-				  const char **result);
+typedef gpg_error_t (*EditAction) (int state, void *opaque,
+				     char **result);
 /* Prototype of the transit function. Returns the next state. If and error
  * is found changes *err. If there is no error it should NOT touch it */
-typedef int (*EditTransit) (int current_state, GpgmeStatusCode status, 
-			    const char *args, void *opaque, GpgmeError *err);
+typedef int (*EditTransit) (int current_state, gpgme_status_code_t status, 
+			    const char *args, void *opaque, gpg_error_t *err);
 
 /* Data to be passed to the edit callback. Must be filled by the caller of
  * gpgme_op_edit()  */
@@ -53,21 +55,26 @@ struct edit_parms_s
    * callback does not abort the edit operation, so we must remember any
    * error. In other words, errors from action() or transit() are sticky.
    */
-  GpgmeError err;
+  gpg_error_t err;
   /* The action function */
   EditAction action;
   /* The transit function */
   EditTransit transit;
+  /* The output data object */
+  gpgme_data_t out;
+  /* Signal attachment id */
+  gulong signal_id;
   /* Data to be passed to the previous functions */
   void *opaque;
 };
 
 /* The edit callback proper */
-static GpgmeError
-edit_fnc (void *opaque, GpgmeStatusCode status, const char *args,
-	  const char **result)
+static gpg_error_t
+edit_fnc (void *opaque, gpgme_status_code_t status,
+	  const char *args, int fd)
 {
   struct edit_parms_s *parms = opaque;
+  char *result = NULL;
 
   /* Ignore these status lines, as they don't require any response */
   if (status == GPGME_STATUS_EOF ||
@@ -86,15 +93,21 @@ edit_fnc (void *opaque, GpgmeStatusCode status, const char *args,
   /* Choose the next state based on the current one and the input */
   parms->state = parms->transit (parms->state, status, args, parms->opaque, 
 				 &parms->err);
-  if (parms->err == GPGME_No_Error)
+  if (gpg_err_code (parms->err) == GPG_ERR_NO_ERROR)
     {
-      GpgmeError err;
+      gpg_error_t err;
       /* Choose the action based on the state */
-      err = parms->action (parms->state, parms->opaque, result);
-      if (err != GPGME_No_Error)
+      err = parms->action (parms->state, parms->opaque, &result);
+      if (gpg_err_code (err) != GPG_ERR_NO_ERROR)
 	{
 	  parms->err = err;
 	}
+      /* Send the command, if any was provided */
+      if (result)
+	{
+	  write (fd, result, strlen (result));
+	  write (fd, "\n", 1);
+	}
     }
   return parms->err;
 }
@@ -112,10 +125,10 @@ typedef enum
   EXPIRE_ERROR
 } ExpireState;
 
-static GpgmeError
-edit_expire_fnc_action (int state, void *opaque, const char **result)
+static gpg_error_t
+edit_expire_fnc_action (int state, void *opaque, char **result)
 {
-  const gchar *date = opaque;
+  gchar *date = opaque;
   
   switch (state)
     {
@@ -140,14 +153,14 @@ edit_expire_fnc_action (int state, void *opaque, const char **result)
       break;
       /* Can't happen */
     default:
-      return GPGME_General_Error;
+      return gpg_error (GPG_ERR_GENERAL);
     }
-  return GPGME_No_Error;
+  return gpg_error (GPG_ERR_NO_ERROR);
 }
 
 static int
-edit_expire_fnc_transit (int current_state, GpgmeStatusCode status, 
-			 const char *args, void *opaque, GpgmeError *err)
+edit_expire_fnc_transit (int current_state, gpgme_status_code_t status, 
+			 const char *args, void *opaque, gpg_error_t *err)
 {
   int next_state;
  
@@ -162,7 +175,7 @@ edit_expire_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = EXPIRE_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case EXPIRE_COMMAND:
@@ -174,7 +187,7 @@ edit_expire_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = EXPIRE_ERROR;
-          *err = GPGME_General_Error;
+          *err =  gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case EXPIRE_DATE:
@@ -186,7 +199,7 @@ edit_expire_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = EXPIRE_ERROR;
-          *err = GPGME_General_Error;
+          *err =  gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case EXPIRE_QUIT:
@@ -198,7 +211,7 @@ edit_expire_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = EXPIRE_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case EXPIRE_ERROR:
@@ -215,7 +228,7 @@ edit_expire_fnc_transit (int current_state, GpgmeStatusCode status,
       break;
     default:
       next_state = EXPIRE_ERROR;
-      *err = GPGME_General_Error;
+      *err = gpg_error (GPG_ERR_GENERAL);
     }
   return next_state;
 }
@@ -234,10 +247,10 @@ typedef enum
   TRUST_ERROR
 } TrustState;
 
-static GpgmeError
-edit_trust_fnc_action (int state, void *opaque, const char **result)
+static gpg_error_t
+edit_trust_fnc_action (int state, void *opaque, char **result)
 {
-  const gchar *trust = opaque;
+  gchar *trust = opaque;
   
   switch (state)
     {
@@ -266,14 +279,14 @@ edit_trust_fnc_action (int state, void *opaque, const char **result)
       break;
       /* Can't happen */
     default:
-      return GPGME_General_Error;
+      return gpg_error (GPG_ERR_GENERAL);
     }
-  return GPGME_No_Error;
+  return gpg_error (GPG_ERR_NO_ERROR);
 }
 
 static int
-edit_trust_fnc_transit (int current_state, GpgmeStatusCode status, 
-			const char *args, void *opaque, GpgmeError *err)
+edit_trust_fnc_transit (int current_state, gpgme_status_code_t status, 
+			const char *args, void *opaque, gpg_error_t *err)
 {
   int next_state;
 
@@ -288,7 +301,7 @@ edit_trust_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = TRUST_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case TRUST_COMMAND:
@@ -300,7 +313,7 @@ edit_trust_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = TRUST_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case TRUST_VALUE:
@@ -317,7 +330,7 @@ edit_trust_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = TRUST_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case TRUST_REALLY_ULTIMATE:
@@ -329,7 +342,7 @@ edit_trust_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = TRUST_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case TRUST_QUIT:
@@ -341,7 +354,7 @@ edit_trust_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = TRUST_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case TRUST_ERROR:
@@ -358,7 +371,7 @@ edit_trust_fnc_transit (int current_state, GpgmeStatusCode status,
       break;
     default:
       next_state = TRUST_ERROR;
-      *err = GPGME_General_Error;
+      *err = gpg_error (GPG_ERR_GENERAL);
     }
 
   return next_state;
@@ -386,8 +399,8 @@ struct sign_parms_s
   gboolean local;
 };
 
-static GpgmeError
-edit_sign_fnc_action (int state, void *opaque, const char **result)
+static gpg_error_t
+edit_sign_fnc_action (int state, void *opaque, char **result)
 {
   struct sign_parms_s *parms = opaque;
   
@@ -426,14 +439,14 @@ edit_sign_fnc_action (int state, void *opaque, const char **result)
       break;
       /* Can't happen */
     default:
-      return GPGME_General_Error;
+      return gpg_error (GPG_ERR_GENERAL);
     }
-  return GPGME_No_Error;
+  return gpg_error (GPG_ERR_NO_ERROR);
 }
 
 static int
-edit_sign_fnc_transit (int current_state, GpgmeStatusCode status, 
-		       const char *args, void *opaque, GpgmeError *err)
+edit_sign_fnc_transit (int current_state, gpgme_status_code_t status, 
+		       const char *args, void *opaque, gpg_error_t *err)
 {
   int next_state;
 
@@ -448,7 +461,7 @@ edit_sign_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = SIGN_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case SIGN_COMMAND:
@@ -472,19 +485,19 @@ edit_sign_fnc_transit (int current_state, GpgmeStatusCode status,
         {
           /* The key has already been signed with this key */
           next_state = SIGN_ERROR;
-          *err = GPGME_Conflict; 
+          *err =  gpg_error (GPG_ERR_CONFLICT);
         }
       else if (status == GPGME_STATUS_GET_LINE &&
           g_str_equal (args, "keyedit.prompt"))
         {
           /* Failed sign: expired key */
           next_state = SIGN_ERROR;
-          *err = GPGME_Invalid_Key;
+          *err = gpg_error (GPG_ERR_UNUSABLE_PUBKEY);
         }
       else
         {
           next_state = SIGN_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case SIGN_UIDS:
@@ -504,12 +517,12 @@ edit_sign_fnc_transit (int current_state, GpgmeStatusCode status,
         {
           /* Failed sign: expired key */
           next_state = SIGN_ERROR;
-          *err = GPGME_Invalid_Key;
+          *err = gpg_error (GPG_ERR_UNUSABLE_PUBKEY);
         }
       else
         {
           next_state = SIGN_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case SIGN_SET_EXPIRE:
@@ -521,7 +534,7 @@ edit_sign_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = SIGN_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case SIGN_SET_CHECK_LEVEL:
@@ -533,7 +546,7 @@ edit_sign_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = SIGN_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case SIGN_CONFIRM:
@@ -545,7 +558,7 @@ edit_sign_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = SIGN_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case SIGN_QUIT:
@@ -557,7 +570,7 @@ edit_sign_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = SIGN_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case SIGN_ERROR:
@@ -574,7 +587,7 @@ edit_sign_fnc_transit (int current_state, GpgmeStatusCode status,
       break;
     default:
       next_state = SIGN_ERROR;
-      *err = GPGME_General_Error;
+      *err = gpg_error (GPG_ERR_GENERAL);
     }
 
   return next_state;
@@ -591,8 +604,15 @@ typedef enum
   PASSWD_ERROR
 } PasswdState;
 
-static GpgmeError
-edit_passwd_fnc_action (int state, void *opaque, const char **result)
+struct passwd_parms_s 
+{
+  gpgme_passphrase_cb_t func;
+  void *opaque;
+  int request_count;
+};
+
+static gpg_error_t
+edit_passwd_fnc_action (int state, void *opaque, char **result)
 {
   switch (state)
     {
@@ -613,14 +633,14 @@ edit_passwd_fnc_action (int state, void *opaque, const char **result)
       break;
       /* Can't happen */
     default:
-      return GPGME_General_Error;
+      return gpg_error (GPG_ERR_GENERAL);
     }
-  return GPGME_No_Error;
+  return gpg_error (GPG_ERR_NO_ERROR);
 }
 
 static int
-edit_passwd_fnc_transit (int current_state, GpgmeStatusCode status, 
-			 const char *args, void *opaque, GpgmeError *err)
+edit_passwd_fnc_transit (int current_state, gpgme_status_code_t status, 
+			 const char *args, void *opaque, gpg_error_t *err)
 {
   int next_state;
  
@@ -635,7 +655,7 @@ edit_passwd_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = PASSWD_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case PASSWD_COMMAND:
@@ -647,7 +667,7 @@ edit_passwd_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = PASSWD_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case PASSWD_QUIT:
@@ -659,7 +679,7 @@ edit_passwd_fnc_transit (int current_state, GpgmeStatusCode status,
       else
         {
           next_state = PASSWD_ERROR;
-          *err = GPGME_General_Error;
+          *err = gpg_error (GPG_ERR_GENERAL);
         }
       break;
     case PASSWD_ERROR:
@@ -676,92 +696,211 @@ edit_passwd_fnc_transit (int current_state, GpgmeStatusCode status,
       break;
     default:
       next_state = PASSWD_ERROR;
-      *err = GPGME_General_Error;
+      *err = gpg_error (GPG_ERR_GENERAL);
     }
   return next_state;
 }
 
+/* Release the edit parameters needed for setting owner trust. The prototype
+ * is that of a GpaContext's "done" signal handler.
+ */
+static void
+gpa_gpgme_edit_trust_parms_release (GpaContext *ctx, gpg_error_t err,
+				   struct edit_parms_s* parms)
+{
+  gpgme_data_release (parms->out);
+  if (parms->signal_id != 0)
+    {
+      /* Don't run this signal handler again if the context is reused */
+      g_signal_handler_disconnect (ctx, parms->signal_id);
+    }
+  g_free (parms->opaque);
+  g_free (parms);
+}
+
+/* Generate the edit parameters needed for setting owner trust.
+ */
+static struct edit_parms_s*
+gpa_gpgme_edit_trust_parms_new (GpaContext *ctx, const char *trust_string,
+				gpgme_data_t out)
+{
+  struct edit_parms_s *edit_parms = g_malloc (sizeof (struct edit_parms_s));
+
+  edit_parms->state = TRUST_START;
+  edit_parms->err = gpg_error (GPG_ERR_NO_ERROR);
+  edit_parms->action = edit_trust_fnc_action;
+  edit_parms->transit = edit_trust_fnc_transit;
+  edit_parms->signal_id = 0;
+  edit_parms->out = out;
+  edit_parms->opaque = g_strdup (trust_string);
+
+  /* Make sure the cleanup is run when the edit completes */
+  edit_parms->signal_id =
+    g_signal_connect (G_OBJECT (ctx), "done", 
+		      G_CALLBACK (gpa_gpgme_edit_trust_parms_release),
+		      edit_parms);
+
+  return edit_parms;
+}
+
 /* Change the ownertrust of a key */
-GpgmeError gpa_gpgme_edit_trust (GpgmeKey key, GpgmeValidity ownertrust)
+gpg_error_t gpa_gpgme_edit_trust_start (GpaContext *ctx, gpgme_key_t key,
+					gpgme_validity_t ownertrust)
 {
   const gchar *trust_strings[] = {"1", "1", "2", "3", "4", "5"};
-  struct edit_parms_s parms = {TRUST_START, GPGME_No_Error, 
-			       edit_trust_fnc_action, edit_trust_fnc_transit,
-			       (char*) trust_strings[ownertrust]};
-  GpgmeError err;
-  GpgmeData out = NULL;
+  struct edit_parms_s *parms = NULL;
+  gpg_error_t err;
+  gpgme_data_t out = NULL;
 
   err = gpgme_data_new (&out);
-  if (err != GPGME_No_Error)
+  if (gpg_err_code (err) != GPG_ERR_NO_ERROR)
     {
       return err;
     }
-  err = gpgme_op_edit (ctx, key, edit_fnc, &parms, out);
-  gpgme_data_release (out);
+  parms = gpa_gpgme_edit_trust_parms_new (ctx, trust_strings[ownertrust], out);
+  err = gpgme_op_edit_start (ctx->ctx, key, edit_fnc, parms, out);
   return err;
 }
 
-/* Change the expire date of a key */
-GpgmeError gpa_gpgme_edit_expire (GpgmeKey key, GDate *date)
+/* Release the edit parameters needed for changing the expiry date. The 
+ * prototype is that of a GpaContext's "done" signal handler.
+ */
+static void
+gpa_gpgme_edit_expire_parms_release (GpaContext *ctx, gpg_error_t err,
+				     struct edit_parms_s* parms)
 {
-  gchar buf[12];
-  struct edit_parms_s parms = {EXPIRE_START, GPGME_No_Error, 
-			       edit_expire_fnc_action, edit_expire_fnc_transit,
-			       buf};
-  GpgmeError err;
-  GpgmeData out = NULL;
-
-  err = gpgme_data_new (&out);
-  if (err != GPGME_No_Error)
+  gpgme_data_release (parms->out);
+  if (parms->signal_id != 0)
     {
-      return err;
+      /* Don't run this signal handler again if the context is reused */
+      g_signal_handler_disconnect (ctx, parms->signal_id);
     }
+  g_free (parms->opaque);
+  g_free (parms);
+}
+
+/* Generate the edit parameters needed for changing the expiry date.
+ */
+static struct edit_parms_s*
+gpa_gpgme_edit_expire_parms_new (GpaContext *ctx, GDate *date,
+				 gpgme_data_t out)
+{
+  struct edit_parms_s *edit_parms = g_malloc (sizeof (struct edit_parms_s));
+  int buf_len = 12;
+  gchar *buf = g_malloc (buf_len);
+
+  edit_parms->state = EXPIRE_START;
+  edit_parms->err = gpg_error (GPG_ERR_NO_ERROR);
+  edit_parms->action = edit_expire_fnc_action;
+  edit_parms->transit = edit_expire_fnc_transit;
+  edit_parms->signal_id = 0;
+  edit_parms->out = out;
+  edit_parms->opaque = buf;
+
   /* The new expiration date */
   if (date)
     {
-      g_date_strftime (buf, sizeof(buf), "%F", date);
+      g_date_strftime (buf, buf_len, "%F", date);
     }
   else
     {
-      strncpy (buf, "0", sizeof (buf));
+      strncpy (buf, "0", buf_len);
     }
-  err = gpgme_op_edit (ctx, key, edit_fnc, &parms, out);
-  gpgme_data_release (out);
+
+  /* Make sure the cleanup is run when the edit completes */
+  edit_parms->signal_id = 
+    g_signal_connect (G_OBJECT (ctx), "done", 
+		      G_CALLBACK (gpa_gpgme_edit_expire_parms_release),
+		      edit_parms);
+
+  return edit_parms;
+}
+
+/* Change the expire date of a key */
+gpg_error_t gpa_gpgme_edit_expire_start (GpaContext *ctx, gpgme_key_t key,
+					 GDate *date)
+{
+  struct edit_parms_s *parms;
+  gpg_error_t err;
+  gpgme_data_t out = NULL;
+
+  err = gpgme_data_new (&out);
+  if (gpg_err_code (err) != GPG_ERR_NO_ERROR)
+    {
+      return err;
+    }
+
+  parms = gpa_gpgme_edit_expire_parms_new (ctx, date, out);
+  err = gpgme_op_edit_start (ctx->ctx, key, edit_fnc, parms, out);
   return err;
 }
 
-/* Sign this key with the given private key */
-GpgmeError gpa_gpgme_edit_sign (GpgmeKey key, const gchar *private_key_fpr,
-                                gboolean local)
+/* Release the edit parameters needed for signing. The prototype is that of
+ * a GpaContext's "done" signal handler.
+ */
+static void
+gpa_gpgme_edit_sign_parms_release (GpaContext *ctx, gpg_error_t err,
+				   struct edit_parms_s* parms)
 {
-  struct sign_parms_s sign_parms = {"0", local};
-  struct edit_parms_s parms = {SIGN_START, GPGME_No_Error,
-			       edit_sign_fnc_action, edit_sign_fnc_transit,
-			       &sign_parms};
-  GpgmeKey secret_key = gpa_keytable_secret_lookup (keytable, private_key_fpr);
-  GpgmeError err = GPGME_No_Error;
-  GpgmeData out;
-
-  /* This return code is a bit weird, but we must use one :-) */
-  if (!secret_key)
+  gpgme_data_release (parms->out);
+  if (parms->signal_id != 0)
     {
-      return GPGME_No_Recipients;
+      g_signal_handler_disconnect (ctx, parms->signal_id);
     }
+  g_free (parms->opaque);
+  g_free (parms);
+}
+
+/* Generate the edit parameters needed for signing 
+ */
+static struct edit_parms_s*
+gpa_gpgme_edit_sign_parms_new (GpaContext *ctx, char *check_level,
+			       gboolean local, gpgme_data_t out)
+{
+  struct sign_parms_s *sign_parms = g_malloc (sizeof (struct sign_parms_s));
+  struct edit_parms_s *edit_parms = g_malloc (sizeof (struct edit_parms_s));
+
+  edit_parms->state = SIGN_START;
+  edit_parms->err = gpg_error (GPG_ERR_NO_ERROR);
+  edit_parms->action = edit_sign_fnc_action;
+  edit_parms->transit = edit_sign_fnc_transit;
+  edit_parms->signal_id = 0;
+  edit_parms->out = out;
+  edit_parms->opaque = sign_parms;
+  sign_parms->check_level = check_level;
+  sign_parms->local = local;
+
+  /* Make sure the cleanup is run when the edit completes */
+  edit_parms->signal_id =
+    g_signal_connect (G_OBJECT (ctx), "done", 
+		      G_CALLBACK (gpa_gpgme_edit_sign_parms_release),
+		      edit_parms);
+
+  return edit_parms;
+}
+
+/* Sign this key with the given private key */
+gpg_error_t gpa_gpgme_edit_sign_start (GpaContext *ctx, gpgme_key_t key,
+				       gpgme_key_t secret_key, gboolean local)
+{
+  struct edit_parms_s *parms = NULL;
+  gpg_error_t err = gpg_error (GPG_ERR_NO_ERROR);
+  gpgme_data_t out;
 
   err = gpgme_data_new (&out);
-  if (err != GPGME_No_Error)
+  if (gpg_err_code (err) != GPG_ERR_NO_ERROR)
     {
       return err;
     }  
-  gpgme_signers_clear (ctx);
-  err = gpgme_signers_add (ctx, secret_key);
-  if (err != GPGME_No_Error)
+  gpgme_signers_clear (ctx->ctx);
+  err = gpgme_signers_add (ctx->ctx, secret_key);
+  if (gpg_err_code (err) != GPG_ERR_NO_ERROR)
     {
       return err;
     }
-  err = gpgme_op_edit (ctx, key, edit_fnc, &parms, out);
-  gpgme_data_release (out);
-  
+  parms = gpa_gpgme_edit_sign_parms_new (ctx, "0", local, out);
+  err = gpgme_op_edit_start (ctx->ctx, key, edit_fnc, parms, out);
+
   return err;
 }
 
@@ -769,48 +908,103 @@ GpgmeError gpa_gpgme_edit_sign (GpgmeKey key, const gchar *private_key_fpr,
  */
 
 /* Special passphrase callback for use within the passwd command */
-static const char *
-passwd_passphrase_cb (void *opaque, const char *desc, void **r_hd)
+gpg_error_t passwd_passphrase_cb (void *hook, const char *uid_hint,
+				  const char *passphrase_info, 
+				  int prev_was_bad, int fd)
 {
-  int *i = opaque;
+  int *i = hook;
 
-  if (desc && desc[0] == 'E') /* It's "ENTER", so it's the beginning of a
-                               * passphrase question */
+  if (!prev_was_bad) /* It's the beginning of a passphrase question */
     {
       (*i)++;
     }
   
   if (*i == 1)
     {
-      return gpa_passphrase_cb (opaque, desc, r_hd);
+      return gpa_passphrase_cb (hook, uid_hint, passphrase_info, 
+				prev_was_bad, fd);
     }
   else
     {
-      return gpa_change_passphrase_dialog_run (opaque, desc, r_hd);
+      return gpa_change_passphrase_dialog_run (hook, uid_hint, 
+					       passphrase_info, 
+					       prev_was_bad, fd);
     }
 }
 
 
-GpgmeError gpa_gpgme_edit_passwd (GpgmeKey key)
+/* Release the edit parameters needed for changing the passphrase. The 
+ * prototype is that of a GpaContext's "done" signal handler.
+ */
+static void
+gpa_gpgme_edit_passwd_parms_release (GpaContext *ctx, gpg_error_t err,
+				     struct edit_parms_s* parms)
 {
-  struct edit_parms_s parms = {PASSWD_START, GPGME_No_Error,
-			       edit_passwd_fnc_action, edit_passwd_fnc_transit,
-			       (gchar *) ""};
-  GpgmeError err;
-  GpgmeData out;
-  int i = 0;
+  struct passwd_parms_s *passwd_parms = parms->opaque;
+
+  gpgme_data_release (parms->out);
+
+  /* Make sure the normal passphrase callback is used */
+  gpgme_set_passphrase_cb (ctx->ctx, passwd_parms->func, passwd_parms->opaque);
+
+  if (parms->signal_id != 0)
+    {
+      /* Don't run this signal handler again if the context is reused */
+      g_signal_handler_disconnect (ctx, parms->signal_id);
+    }
+  
+  g_free (parms);
+  g_free (parms->opaque);
+}
+
+/* Generate the edit parameters needed for changing the passphrase.
+ */
+static struct edit_parms_s*
+gpa_gpgme_edit_passwd_parms_new (GpaContext *ctx, gpgme_data_t out)
+{
+  struct edit_parms_s *edit_parms = g_malloc (sizeof (struct edit_parms_s));
+  struct passwd_parms_s *passwd_parms = g_malloc (sizeof 
+						  (struct passwd_parms_s));
+
+  edit_parms->state = PASSWD_START;
+  edit_parms->err = gpg_error (GPG_ERR_NO_ERROR);
+  edit_parms->action = edit_passwd_fnc_action;
+  edit_parms->transit = edit_passwd_fnc_transit;
+  edit_parms->signal_id = 0;
+  edit_parms->out = out;
+  edit_parms->opaque = passwd_parms;
+  passwd_parms->request_count = 0;
+  gpgme_get_passphrase_cb (ctx->ctx, &passwd_parms->func,
+			   &passwd_parms->opaque);
+
+  /* Make sure the cleanup is run when the edit completes */
+  edit_parms->signal_id = 
+    g_signal_connect (G_OBJECT (ctx), "done", 
+		      G_CALLBACK (gpa_gpgme_edit_passwd_parms_release),
+		      edit_parms);
+
+  return edit_parms;
+}
+
+gpg_error_t gpa_gpgme_edit_passwd_start (GpaContext *ctx, gpgme_key_t key)
+{
+  struct edit_parms_s *parms;
+  gpg_error_t err;
+  gpgme_data_t out;
 
   err = gpgme_data_new (&out);
-  if (err != GPGME_No_Error)
+  if (gpg_err_code (err) != GPG_ERR_NO_ERROR)
     {
       return err;
     }
-  /* Use our own passphrase callback */
-  gpgme_set_passphrase_cb (ctx, passwd_passphrase_cb, &i);
-  err = gpgme_op_edit (ctx, key, edit_fnc, &parms, out);
-  gpgme_data_release (out);
-  /* Make sure the normal passphrase callback is used */
-  gpgme_set_passphrase_cb (ctx, gpa_passphrase_cb, NULL);
+  parms = gpa_gpgme_edit_passwd_parms_new (ctx, out);
 
-  return GPGME_No_Error;
+  /* Use our own passphrase callback: the data is a pointer to an int
+   * that counts the passphrase requests received */
+  gpgme_set_passphrase_cb (ctx->ctx, passwd_passphrase_cb,
+			   &(((struct passwd_parms_s*) 
+			      parms->opaque)->request_count));
+  err = gpgme_op_edit_start (ctx->ctx, key, edit_fnc, parms, out);
+
+  return err;
 }
